@@ -43,6 +43,7 @@
 #include <osmocom/pfcp/pfcp_cp_peer.h>
 
 #include <osmocom/hnbgw/hnbgw.h>
+#include <osmocom/hnbgw/hnbgw_pfcp.h>
 #include <osmocom/hnbgw/context_map.h>
 #include <osmocom/hnbgw/ranap_rab_ass.h>
 #include <osmocom/hnbgw/ps_rab_ass_fsm.h>
@@ -61,6 +62,7 @@ static const struct value_string ps_rab_ass_fsm_event_names[] = {
 	OSMO_VALUE_STRING(PS_RAB_ASS_EV_LOCAL_F_TEIDS_RX),
 	OSMO_VALUE_STRING(PS_RAB_ASS_EV_RAB_ASS_RESP),
 	OSMO_VALUE_STRING(PS_RAB_ASS_EV_RAB_ESTABLISHED),
+	OSMO_VALUE_STRING(PS_RAB_ASS_EV_RAB_RELEASED),
 	OSMO_VALUE_STRING(PS_RAB_ASS_EV_RAB_FAIL),
 	{}
 };
@@ -81,10 +83,10 @@ enum ps_rab_ass_state {
  * This structure manages the RAB Assignment procedures, and the currently set up RABs:
  *
  * - hnbgw_context_map
- *   - .ps_rab_ass: list of PS RAB Assignment procedures
+ *   - .ps_rab_ass_list: list of PS RAB Assignment procedures
  *     - ps_rab_ass_fsm: one RANAP PS RAB Assignment procedure
  *     - ...
- *   - .ps_rabs: list of individual PS RABs
+ *   - .ps_rab_list: list of individual PS RABs
  *     - ps_rab_fsm: one GTP mapping with PFCP session to the UPF, for a single RAB
  *     - ...
  *
@@ -111,6 +113,9 @@ struct ps_rab_ass {
 	 * ps_rab_fsms, without iterating the RAB Assignment message every time (minor optimisation). */
 	int rabs_count;
 	int rabs_done_count;
+
+	unsigned int rabs_rel_count;
+	unsigned int rabs_rel_done_count;
 };
 
 struct osmo_tdef_state_timeout ps_rab_ass_fsm_timeouts[32] = {
@@ -118,7 +123,7 @@ struct osmo_tdef_state_timeout ps_rab_ass_fsm_timeouts[32] = {
 	/* PS_RAB_ASS_ST_WAIT_RABS_ESTABLISHED is terminated by PFCP timeouts via ps_rab_fsm */
 };
 
-#define ps_rab_ass_fsm_state_chg(state) \
+#define ps_rab_ass_fsm_state_chg(fi, state) \
 	osmo_tdef_fsm_inst_state_chg(fi, state, ps_rab_ass_fsm_timeouts, hnbgw_T_defs, -1)
 
 static struct osmo_fsm ps_rab_ass_fsm;
@@ -140,7 +145,7 @@ static struct ps_rab_ass *ps_rab_ass_alloc(struct hnbgw_context_map *map)
 	};
 	fi->priv = rab_ass;
 
-	llist_add_tail(&rab_ass->entry, &map->ps_rab_ass);
+	llist_add_tail(&rab_ass->entry, &map->ps_rab_ass_list);
 	return rab_ass;
 }
 
@@ -219,48 +224,51 @@ int hnbgw_gtpmap_rx_rab_ass_req(struct hnbgw_context_map *map, struct msgb *rana
 {
 	RANAP_RAB_AssignmentRequestIEs_t *ies = &message->msg.raB_AssignmentRequestIEs;
 	int i;
-
 	struct ps_rab_ass *rab_ass;
-	struct osmo_fsm_inst *fi;
 
 	rab_ass = ps_rab_ass_alloc(map);
-
 	talloc_steal(rab_ass, message);
 	rab_ass->ranap_rab_ass_req_message = message;
 	/* Now rab_ass owns message and will clean it up */
 
-	if (!osmo_pfcp_cp_peer_is_associated(g_hnbgw->pfcp.cp_peer)) {
+	if (!osmo_pfcp_cp_peer_is_associated(g_hnbgw->pfcp.upf->cp_peer)) {
 		LOG_MAP(map, DLPFCP, LOGL_ERROR, "PFCP is not associated, cannot set up GTP mapping\n");
 		goto no_rab;
 	}
 
-	/* Make sure we indeed deal with a setup-or-modify list */
-	if (!(ies->presenceMask & RAB_ASSIGNMENTREQUESTIES_RANAP_RAB_SETUPORMODIFYLIST_PRESENT)) {
-		LOG_MAP(map, DLPFCP, LOGL_ERROR, "RANAP PS RAB AssignmentRequest lacks setup-or-modify list\n");
-		goto no_rab;
+	/* setup-or-modify list */
+	if (ies->presenceMask & RAB_ASSIGNMENTREQUESTIES_RANAP_RAB_SETUPORMODIFYLIST_PRESENT) {
+		/* Multiple RABs may be set up, assemble in list map->ps_rab_list. */
+		for (i = 0; i < ies->raB_SetupOrModifyList.list.count; i++) {
+			RANAP_ProtocolIE_ContainerPair_t *protocol_ie_container_pair;
+			RANAP_ProtocolIE_FieldPair_t *protocol_ie_field_pair;
+
+			protocol_ie_container_pair = ies->raB_SetupOrModifyList.list.array[i];
+			protocol_ie_field_pair = protocol_ie_container_pair->list.array[0];
+			if (!protocol_ie_field_pair)
+				goto no_rab;
+			if (protocol_ie_field_pair->id != RANAP_ProtocolIE_ID_id_RAB_SetupOrModifyItem)
+				goto no_rab;
+
+			if (ps_rab_setup_core_remote(rab_ass, protocol_ie_field_pair))
+				goto no_rab;
+		}
 	}
 
-	/* Multiple RABs may be set up, assemble in list rab_ass->ps_rabs. */
-	for (i = 0; i < ies->raB_SetupOrModifyList.list.count; i++) {
-		RANAP_ProtocolIE_ContainerPair_t *protocol_ie_container_pair;
-		RANAP_ProtocolIE_FieldPair_t *protocol_ie_field_pair;
+	/* Note: ReleaseList is handled at RAB Assign Response time. */
 
-		protocol_ie_container_pair = ies->raB_SetupOrModifyList.list.array[i];
-		protocol_ie_field_pair = protocol_ie_container_pair->list.array[0];
-		if (!protocol_ie_field_pair)
-			goto no_rab;
-		if (protocol_ie_field_pair->id != RANAP_ProtocolIE_ID_id_RAB_SetupOrModifyItem)
-			goto no_rab;
-
-		if (ps_rab_setup_core_remote(rab_ass, protocol_ie_field_pair))
-			goto no_rab;
+	if (rab_ass->rabs_count > 0) {
+		/* Got all RABs' state and their Core side GTP info in map->ps_rab_list.
+		 * For each, a ps_rab_fsm has been started and each will call back with
+		 * PS_RAB_ASS_EV_LOCAL_F_TEIDS_RX or PS_RAB_ASS_EV_RAB_FAIL. */
+		return ps_rab_ass_fsm_state_chg(rab_ass->fi, PS_RAB_ASS_ST_WAIT_LOCAL_F_TEIDS);
 	}
 
-	/* Got all RABs' state and their Core side GTP info in map->ps_rabs. For each, a ps_rab_fsm has been started and
-	 * each will call back with PS_RAB_ASS_EV_LOCAL_F_TEIDS_RX or PS_RAB_ASS_EV_RAB_FAIL. */
-	fi = rab_ass->fi;
-	return ps_rab_ass_fsm_state_chg(PS_RAB_ASS_ST_WAIT_LOCAL_F_TEIDS);
-
+	/* No Setup-or-modify, only Release RABs. Forward the message as is,
+	 * RABs will be released at RAB ASS Response time. */
+	map_rua_dispatch(map, MAP_RUA_EV_TX_DIRECT_TRANSFER, ranap_msg);
+	osmo_fsm_inst_term(rab_ass->fi, OSMO_FSM_TERM_REGULAR, NULL);
+	return 0;
 no_rab:
 	ps_rab_ass_failure(rab_ass);
 	return -1;
@@ -413,70 +421,127 @@ int hnbgw_gtpmap_rx_rab_ass_resp(struct hnbgw_context_map *map, struct msgb *ran
 	 * each other, we are able to handle mismatching RAB IDs in request and response messages.
 	 */
 
+	RANAP_RAB_AssignmentResponseIEs_t *ies = &message->msg.raB_AssignmentResponseIEs;
 	int rc;
 	int i;
 	struct ps_rab_ass *rab_ass;
-	struct osmo_fsm_inst *fi;
-	RANAP_RAB_AssignmentResponseIEs_t *ies;
-
-	/* Make sure we indeed deal with a setup-or-modify list */
-	ies = &message->msg.raB_AssignmentResponseIEs;
-	if (!(ies->presenceMask & RAB_ASSIGNMENTRESPONSEIES_RANAP_RAB_SETUPORMODIFIEDLIST_PRESENT)) {
-		LOG_MAP(map, DRUA, LOGL_ERROR, "RANAP PS RAB AssignmentResponse lacks setup-or-modify list\n");
-		return -1;
-	}
 
 	rab_ass = ps_rab_ass_alloc(map);
-
 	talloc_steal(rab_ass, message);
 	rab_ass->ranap_rab_ass_resp_message = message;
-
 	talloc_steal(rab_ass, ranap_msg);
 	rab_ass->ranap_rab_ass_resp_msgb = ranap_msg;
 	/* Now rab_ass owns message and will clean it up */
 
-	if (!osmo_pfcp_cp_peer_is_associated(g_hnbgw->pfcp.cp_peer)) {
+	if (!osmo_pfcp_cp_peer_is_associated(g_hnbgw->pfcp.upf->cp_peer)) {
 		LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "PFCP is not associated, cannot set up GTP mapping\n");
-		ps_rab_ass_failure(rab_ass);
-		return -1;
+		goto no_rab;
 	}
 
-	LOG_PS_RAB_ASS(rab_ass, LOGL_NOTICE, "PS RAB-AssignmentResponse received, updating RABs\n");
+	/* setup-or-modify list */
+	if (ies->presenceMask & RAB_ASSIGNMENTRESPONSEIES_RANAP_RAB_SETUPORMODIFIEDLIST_PRESENT) {
 
-	/* Multiple RABs may be set up, bump matching FSMs in list rab_ass->ps_rabs. */
-	for (i = 0; i < ies->raB_SetupOrModifiedList.raB_SetupOrModifiedList_ies.list.count; i++) {
-		RANAP_IE_t *list_ie;
-		RANAP_RAB_SetupOrModifiedItemIEs_t item_ies;
+		LOG_PS_RAB_ASS(rab_ass, LOGL_INFO, "PS RAB-AssignmentResponse received, updating RABs\n");
 
-		list_ie = ies->raB_SetupOrModifiedList.raB_SetupOrModifiedList_ies.list.array[i];
-		if (!list_ie)
-			continue;
+		/* Multiple RABs may be set up, bump matching FSMs in list map->ps_rab_list. */
+		for (i = 0; i < ies->raB_SetupOrModifiedList.raB_SetupOrModifiedList_ies.list.count; i++) {
+			RANAP_IE_t *list_ie;
+			RANAP_RAB_SetupOrModifiedItemIEs_t item_ies;
 
-		rc = ranap_decode_rab_setupormodifieditemies_fromlist(&item_ies,
-								      &list_ie->value);
-		if (rc < 0) {
-			LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Failed to decode PS RAB-AssignmentResponse"
-				       " SetupOrModifiedItemIEs with list index %d\n", i);
-			goto continue_cleanloop;
-		}
+			list_ie = ies->raB_SetupOrModifiedList.raB_SetupOrModifiedList_ies.list.array[i];
+			if (!list_ie)
+				continue;
 
-		if (ps_rab_setup_access_remote(rab_ass, &item_ies.raB_SetupOrModifiedItem))
-			LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Failed to apply PS RAB-AssignmentResponse"
-				       " SetupOrModifiedItemIEs with list index %d\n", i);
-		rab_ass->rabs_count++;
+			rc = ranap_decode_rab_setupormodifieditemies_fromlist(&item_ies,
+									      &list_ie->value);
+			if (rc < 0) {
+				LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Failed to decode PS RAB-AssignmentResponse"
+					       " SetupOrModifiedItemIEs with list index %d\n", i);
+				goto continue_cleanloop;
+			}
+
+			if (ps_rab_setup_access_remote(rab_ass, &item_ies.raB_SetupOrModifiedItem))
+				LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Failed to apply PS RAB-AssignmentResponse"
+					       " SetupOrModifiedItemIEs with list index %d\n", i);
+			rab_ass->rabs_count++;
 
 continue_cleanloop:
-		ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_SetupOrModifiedItem, &item_ies);
+			ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_SetupOrModifiedItem, &item_ies);
+		}
 	}
 
-	/* Got all RABs' state and updated their Access side GTP info in map->ps_rabs. For each RAB ID, the matching
-	 * ps_rab_fsm has been instructed to tell the UPF about the Access Remote GTP F-TEID. Each will call back with
-	 * PS_RAB_ASS_EV_RAB_ESTABLISHED or PS_RAB_ASS_EV_RAB_FAIL. */
-	fi = rab_ass->fi;
-	return ps_rab_ass_fsm_state_chg(PS_RAB_ASS_ST_WAIT_RABS_ESTABLISHED);
+	/* release list */
+	if (ies->presenceMask & RAB_ASSIGNMENTRESPONSEIES_RANAP_RAB_RELEASEDLIST_PRESENT) {
+		/* Multiple RABs may be set up, assemble in list map->ps_rab_list. */
+		RANAP_RAB_ReleasedList_t *r_list = &ies->raB_ReleasedList;
+		for (i = 0; i < r_list->raB_ReleasedList_ies.list.count; i++) {
+			RANAP_IE_t *released_list_ie = r_list->raB_ReleasedList_ies.list.array[i];
+			RANAP_RAB_ReleasedItemIEs_t _rab_rel_item_ies = {};
+			RANAP_RAB_ReleasedItemIEs_t *rab_rel_item_ies = &_rab_rel_item_ies;
+			RANAP_RAB_ReleasedItem_t *rab_rel_item;
+			uint8_t rab_id;
+			struct ps_rab *rab;
+
+			if (!released_list_ie)
+				goto no_rab;
+			if (released_list_ie->id != RANAP_ProtocolIE_ID_id_RAB_ReleasedItem)
+				goto no_rab;
+
+			rc = ranap_decode_rab_releaseditemies_fromlist(rab_rel_item_ies, &released_list_ie->value);
+			if (rc < 0) {
+				LOG_MAP(map, DLPFCP, LOGL_NOTICE, "Rx RAB ASS REQ (REL): Failed decoding ReleaseItem %u\n", i);
+				goto no_rab;
+			}
+
+			rab_rel_item = &rab_rel_item_ies->raB_ReleasedItem;
+			/* The RAB-ID is defined as a bitstring with a size of 8 (1 byte),
+			 * See also RANAP-IEs.asn, RAB-ID ::= BIT STRING (SIZE (8)) */
+			rab_id = rab_rel_item->rAB_ID.buf[0];
+			ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_RANAP_RAB_ReleasedItem, rab_rel_item_ies);
+
+			rab = ps_rab_get(map, rab_id);
+			if (rab) {
+				rab_ass->rabs_rel_count++;
+				ps_rab_release(rab, rab_ass->fi);
+			} else {
+				LOG_MAP(map, DLPFCP, LOGL_NOTICE, "Rx RAB ASS REQ (REL) for unknown rab_id %u\n", rab_id);
+			}
+		}
+	}
+
+	if (rab_ass->rabs_count > 0 || rab_ass->rabs_rel_count > 0) {
+		/* Got all RABs' state and updated their Access side GTP info in map->ps_rab_list.
+		 * For each RAB ID, the matching ps_rab_fsm has been instructed to tell the UPF about
+		 * the Access Remote GTP F-TEID. Each will call back with PS_RAB_ASS_EV_RAB_ESTABLISHED
+		 * or PS_RAB_ASS_EV_RAB_FAIL. */
+		return ps_rab_ass_fsm_state_chg(rab_ass->fi, PS_RAB_ASS_ST_WAIT_RABS_ESTABLISHED);
+	}
+
+	/* No Setup-or-modify nor Release RABs. Forward the message as is. */
+	rc = map_sccp_dispatch(map, MAP_SCCP_EV_TX_DATA_REQUEST, ranap_msg);
+	if (rc < 0) {
+		LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Sending RANAP PS RAB-AssignmentResponse failed\n");
+		goto no_rab;
+	}
+	/* The request message has been forwarded. We are done. */
+	osmo_fsm_inst_term(rab_ass->fi, OSMO_FSM_TERM_REGULAR, NULL);
+	return 0;
+
+no_rab:
+	ps_rab_ass_failure(rab_ass);
+	return -1;
 }
 
 static void ps_rab_ass_resp_send_if_ready(struct ps_rab_ass *rab_ass);
+static void ps_rab_ass_resp_send_if_ready_quick(struct ps_rab_ass *rab_ass)
+{
+	/* some RABs are still pending, postpone going through the message until all are done. */
+	if (rab_ass->rabs_done_count < rab_ass->rabs_count)
+		return;
+	if (rab_ass->rabs_rel_done_count < rab_ass->rabs_rel_count)
+		return;
+	ps_rab_ass_resp_send_if_ready(rab_ass);
+}
 
 static void ps_rab_ass_fsm_wait_rabs_established(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
@@ -485,12 +550,13 @@ static void ps_rab_ass_fsm_wait_rabs_established(struct osmo_fsm_inst *fi, uint3
 	switch (event) {
 	case PS_RAB_ASS_EV_RAB_ESTABLISHED:
 		rab_ass->rabs_done_count++;
-		if (rab_ass->rabs_done_count < rab_ass->rabs_count) {
-			/* some RABs are still pending, postpone going through the message until all are done. */
-			return;
-		}
 		/* All RABs have succeeded, ready to forward */
-		ps_rab_ass_resp_send_if_ready(rab_ass);
+		ps_rab_ass_resp_send_if_ready_quick(rab_ass);
+		return;
+
+	case PS_RAB_ASS_EV_RAB_RELEASED:
+		rab_ass->rabs_rel_done_count++;
+		ps_rab_ass_resp_send_if_ready_quick(rab_ass);
 		return;
 
 	case PS_RAB_ASS_EV_RAB_FAIL:
@@ -601,14 +667,14 @@ continue_cleanloop:
 		return;
 	}
 
-	LOG_PS_RAB_ASS(rab_ass, LOGL_NOTICE, "Sending RANAP PS RAB-AssignmentResponse with mapped GTP info\n");
+	LOG_PS_RAB_ASS(rab_ass, LOGL_INFO, "Sending RANAP PS RAB-AssignmentResponse with mapped GTP info\n");
 
 	rc = map_sccp_dispatch(rab_ass->map, MAP_SCCP_EV_TX_DATA_REQUEST, msg);
 	if (rc < 0) {
 		LOG_PS_RAB_ASS(rab_ass, LOGL_ERROR, "Sending RANAP PS RAB-AssignmentResponse failed\n");
 		ps_rab_ass_failure(rab_ass);
+		return;
 	}
-
 	/* The request message has been forwarded. We are done. */
 	osmo_fsm_inst_term(rab_ass->fi, OSMO_FSM_TERM_REGULAR, NULL);
 }
@@ -618,14 +684,14 @@ static void ps_rab_ass_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_
 	struct ps_rab_ass *rab_ass = fi->priv;
 	struct ps_rab *rab;
 
-	llist_for_each_entry(rab, &rab_ass->map->ps_rabs, entry) {
+	llist_for_each_entry(rab, &rab_ass->map->ps_rab_list, entry) {
 		if (rab->req_fi == fi)
 			rab->req_fi = NULL;
 		if (rab->resp_fi == fi)
 			rab->resp_fi = NULL;
 	}
 
-	/* remove from map->ps_rab_ass */
+	/* remove from map->ps_rab_ass_list */
 	llist_del(&rab_ass->entry);
 }
 
@@ -633,10 +699,10 @@ void hnbgw_gtpmap_release(struct hnbgw_context_map *map)
 {
 	struct ps_rab_ass *rab_ass, *next;
 	struct ps_rab *rab, *next2;
-	llist_for_each_entry_safe(rab, next2, &map->ps_rabs, entry) {
-		ps_rab_release(rab);
+	llist_for_each_entry_safe(rab, next2, &map->ps_rab_list, entry) {
+		ps_rab_release(rab, NULL);
 	}
-	llist_for_each_entry_safe(rab_ass, next, &map->ps_rab_ass, entry) {
+	llist_for_each_entry_safe(rab_ass, next, &map->ps_rab_ass_list, entry) {
 		osmo_fsm_inst_term(rab_ass->fi, OSMO_FSM_TERM_REGULAR, NULL);
 	}
 }
@@ -664,6 +730,7 @@ static const struct osmo_fsm_state ps_rab_ass_fsm_states[] = {
 		.action = ps_rab_ass_fsm_wait_rabs_established,
 		.in_event_mask = 0
 			| S(PS_RAB_ASS_EV_RAB_ESTABLISHED)
+			| S(PS_RAB_ASS_EV_RAB_RELEASED)
 			| S(PS_RAB_ASS_EV_RAB_FAIL)
 			,
 	},
